package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.advancements.CriteriaTriggers
import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.LongTag
import net.minecraft.nbt.StringTag
import net.minecraft.network.chat.Component
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerLevel
import net.minecraft.server.level.ServerPlayer
import net.minecraft.sounds.SoundSource
import net.minecraft.world.Container
import net.minecraft.world.Containers
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.gameevent.GameEvent
import net.minecraft.world.level.storage.loot.LootContext
import net.minecraft.world.level.storage.loot.parameters.LootContextParamSets
import net.minecraft.world.level.storage.loot.parameters.LootContextParams
import net.minecraft.world.phys.Vec3
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.block.CargoCrateBlock
import ru.dbotthepony.mc.otm.block.IDroppableContainer
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.MatteryContainerHooks
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.map
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.menu.CargoCrateMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MSoundEvents

class CargoCrateBlockEntity(
	p_155229_: BlockPos,
	p_155230_: BlockState
) : MatteryBlockEntity(MBlockEntities.CARGO_CRATE, p_155229_, p_155230_), IDroppableContainer {
	val container = MatteryContainer(this::setChanged, CAPACITY)
	private var interactingPlayers = 0
	val handler = container.handler(object : MatteryContainerHooks {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return lootTable == null
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return lootTable == null
		}
	})

	override fun beforeDroppingItems(
		oldBlockState: BlockState,
		level: Level,
		blockPos: BlockPos,
		newBlockState: BlockState,
		movedByPiston: Boolean
	) {
		unpackLootTable()
	}

	override fun beforeDestroyedByPlayer(level: Level, blockPos: BlockPos, blockState: BlockState, player: Player) {
		unpackLootTable(player)
	}

	var lootTable: ResourceLocation? = null
	var lootTableSeed: Long? = null

	override val droppableContainer: Container
		get() = container

	fun onPlayerOpen() {
		val level = level

		if (interactingPlayers++ == 0 && level != null) {
			level.setBlock(blockPos, blockState.setValue(CargoCrateBlock.IS_OPEN, true), Block.UPDATE_CLIENTS)
			level.playSound(null, blockPos, MSoundEvents.CARGO_CRATE_OPEN, SoundSource.BLOCKS, 1f, 0.8f + level.random.nextFloat() * 0.2f)
			level.gameEvent(GameEvent.CONTAINER_OPEN, blockPos, GameEvent.Context.of(blockState))
		}
	}

	fun onPlayerClose() {
		val level = level

		if (--interactingPlayers == 0 && level != null) {
			level.setBlock(blockPos, blockState.setValue(CargoCrateBlock.IS_OPEN, false), Block.UPDATE_CLIENTS)
			level.gameEvent(GameEvent.CONTAINER_CLOSE, blockPos, GameEvent.Context.of(blockState))
		}
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		handler.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		handler.revive()
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (cap == ForgeCapabilities.ITEM_HANDLER)
			return handler.get().cast()

		return super.getCapability(cap, side)
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt[INVENTORY_KEY] = container.serializeNBT()

		lootTable?.let { nbt[LOOT_TABLE_KEY] = it.toString() }
		lootTableSeed?.let { nbt[LOOT_TABLE_SEED_KEY] = it }
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		container.deserializeNBT(nbt[INVENTORY_KEY])

		lootTable = nbt.map(LOOT_TABLE_KEY) { it: StringTag -> ResourceLocation.tryParse(it.asString) }
		lootTableSeed = (nbt[LOOT_TABLE_SEED_KEY] as LongTag?)?.asLong
	}

	fun unpackLootTable(ply: Player? = null) {
		val lootTable = lootTable ?: return
		val lootTableSeed = lootTableSeed ?: 0L
		val server = level?.server ?: return

		val loot = server.lootTables.get(lootTable)

		if (ply is ServerPlayer) {
			CriteriaTriggers.GENERATE_LOOT.trigger(ply, lootTable)
		}

		this.lootTable = null
		this.lootTableSeed = null

		val context = LootContext.Builder(level as ServerLevel)
			.withParameter(LootContextParams.ORIGIN, Vec3.atCenterOf(this.worldPosition))
			.withOptionalRandomSeed(lootTableSeed)

		if (ply != null) {
			context.withLuck(ply.luck).withParameter(LootContextParams.THIS_ENTITY, ply)
		}

		Containers.dropContents(level as ServerLevel, blockPos, container)
		loot.fill(container, context.create(LootContextParamSets.CHEST))
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu? {
		if (lootTable != null && ply.isSpectator)
			return null

		unpackLootTable(ply)
		return CargoCrateMenu(containerID, inventory, this)
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.cargo_crate")
		const val CAPACITY = 9 * 6
	}
}
