
@file:Suppress("unused")

package ru.dbotthepony.mc.otm.core

import com.google.common.collect.ImmutableList
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import net.minecraft.core.BlockPos
import net.minecraft.nbt.ByteArrayTag
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.Tag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.entity.Entity
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.crafting.Ingredient
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.block.state.StateHolder
import net.minecraft.world.level.block.state.properties.Property
import net.minecraft.world.phys.Vec3
import net.minecraftforge.common.ForgeHooks
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.IItemHandler
import net.minecraftforge.registries.ForgeRegistries
import net.minecraftforge.registries.ForgeRegistry
import net.minecraftforge.registries.IForgeRegistry
import java.math.BigInteger
import java.util.Arrays
import java.util.Spliterators
import java.util.UUID
import java.util.stream.Stream
import java.util.stream.StreamSupport
import kotlin.reflect.KProperty

/**
 * Performs type check+cast and sums two numbers.
 *
 * Returns the same type as on input (adding Double to Float will not return Double, etc.)
 */
fun Number.dynPlus(other: Number): Number {
	return when (this) {
		is Float -> this 	+ other.toFloat()
		is Double -> this 	+ other.toDouble()
		is Int -> this 		+ other.toInt()
		is Long -> this 	+ other.toLong()
		is Short -> this 	+ other.toShort()
		is Byte -> this 	+ other.toByte()

		is ImpreciseFraction -> {
			when (other) {
				is Float -> this 					+ other.toFloat()
				is Double -> this 					+ other.toDouble()
				is Int, is Byte, is Short -> this 	+ other.toInt()
				is Long -> this 					+ other.toLong()
				is ImpreciseFraction -> this 		+ other
				is Fraction -> this 				+ other.toImpreciseFraction()
				else -> throw IllegalStateException("I don't know what to do with ${other::class.qualifiedName} (right hand operand)")
			}
		}

		is Fraction -> {
			when (other) {
				is Float -> this 						+ other.toFloat()
				is Double -> this 						+ other.toDouble()
				is Int, is Byte, is Short -> this 		+ other.toInt()
				is Long -> this 						+ other.toLong()
				is ImpreciseFraction -> this 			+ other.toFraction()
				is Fraction -> this 					+ other
				else -> throw IllegalStateException("I don't know what to do with ${other::class.qualifiedName} (right hand operand)")
			}
		}

		else -> throw IllegalStateException("I don't know what to do with ${this::class.qualifiedName} (left hand operand)")
	}
}

/**
 * Performs type check+cast and subtracts two numbers.
 *
 * Returns the same type as on input (adding Double to Float will not return Double, etc.)
 */
fun Number.dynMinus(other: Number): Number {
	return when (this) {
		is Float -> this 	- other.toFloat()
		is Double -> this 	- other.toDouble()
		is Int -> this 		- other.toInt()
		is Long -> this 	- other.toLong()
		is Short -> this 	- other.toShort()
		is Byte -> this 	- other.toByte()

		is ImpreciseFraction -> {
			when (other) {
				is Float -> this 					- other.toFloat()
				is Double -> this 					- other.toDouble()
				is Int, is Byte, is Short -> this 	- other.toInt()
				is Long -> this 					- other.toLong()
				is ImpreciseFraction -> this 		- other
				is Fraction -> this 				- other.toImpreciseFraction()
				else -> throw IllegalStateException("I don't know what to do with ${other::class.qualifiedName} (right hand operand)")
			}
		}

		is Fraction -> {
			when (other) {
				is Float -> this 						- other.toFloat()
				is Double -> this 						- other.toDouble()
				is Int, is Byte, is Short -> this 		- other.toInt()
				is Long -> this 						- other.toLong()
				is ImpreciseFraction -> this 			- other.toFraction()
				is Fraction -> this 					- other
				else -> throw IllegalStateException("I don't know what to do with ${other::class.qualifiedName} (right hand operand)")
			}
		}

		else -> throw IllegalStateException("I don't know what to do with ${this::class.qualifiedName} (left hand operand)")
	}
}

/**
 * Performs type check+cast and multiplies two numbers.
 *
 * Returns the same type as on input (adding Double to Float will not return Double, etc.)
 */
fun Number.dynTimes(other: Number): Number {
	return when (this) {
		is Float -> this 	* other.toFloat()
		is Double -> this 	* other.toDouble()
		is Int -> this 		* other.toInt()
		is Long -> this 	* other.toLong()
		is Short -> this 	* other.toShort()
		is Byte -> this 	* other.toByte()

		is ImpreciseFraction -> {
			when (other) {
				is Float -> this 					* other.toFloat()
				is Double -> this 					* other.toDouble()
				is Int, is Byte, is Short -> this 	* other.toInt()
				is Long -> this 					* other.toLong()
				is ImpreciseFraction -> this 		* other
				is Fraction -> this 				* other.toImpreciseFraction()
				else -> throw IllegalStateException("I don't know what to do with ${other::class.qualifiedName} (right hand operand)")
			}
		}

		is Fraction -> {
			when (other) {
				is Float -> this 						* other.toFloat()
				is Double -> this 						* other.toDouble()
				is Int, is Byte, is Short -> this 		* other.toInt()
				is Long -> this 						* other.toLong()
				is ImpreciseFraction -> this 			* other.toFraction()
				is Fraction -> this 					* other
				else -> throw IllegalStateException("I don't know what to do with ${other::class.qualifiedName} (right hand operand)")
			}
		}

		else -> throw IllegalStateException("I don't know what to do with ${this::class.qualifiedName} (left hand operand)")
	}
}

/**
 * Performs type check+cast and divides two numbers.
 *
 * Returns the same type as on input (adding Double to Float will not return Double, etc.)
 */
fun Number.dynDiv(other: Number): Number {
	return when (this) {
		is Float -> this 	/ other.toFloat()
		is Double -> this 	/ other.toDouble()
		is Int -> this 		/ other.toInt()
		is Long -> this 	/ other.toLong()
		is Short -> this 	/ other.toShort()
		is Byte -> this 	/ other.toByte()

		is ImpreciseFraction -> {
			when (other) {
				is Float -> this 					/ other.toFloat()
				is Double -> this 					/ other.toDouble()
				is Int, is Byte, is Short -> this 	/ other.toInt()
				is Long -> this 					/ other.toLong()
				is ImpreciseFraction -> this 		/ other
				is Fraction -> this 				/ other.toImpreciseFraction()
				else -> throw IllegalStateException("I don't know what to do with ${other::class.qualifiedName} (right hand operand)")
			}
		}

		is Fraction -> {
			when (other) {
				is Float -> this 						/ other.toFloat()
				is Double -> this 						/ other.toDouble()
				is Int, is Byte, is Short -> this 		/ other.toInt()
				is Long -> this 						/ other.toLong()
				is ImpreciseFraction -> this 			/ other.toFraction()
				is Fraction -> this 					/ other
				else -> throw IllegalStateException("I don't know what to do with ${other::class.qualifiedName} (right hand operand)")
			}
		}

		else -> throw IllegalStateException("I don't know what to do with ${this::class.qualifiedName} (left hand operand)")
	}
}

val Number.isFractional get() = this is Float || this is Double || this is ImpreciseFraction || this is Fraction
val Number.isWhole get() = !isFractional

fun Number.toImpreciseFraction(): ImpreciseFraction {
	return when (this) {
		is ImpreciseFraction -> this
		is Float -> ImpreciseFraction(this)
		is Double -> ImpreciseFraction(this)
		is Int -> ImpreciseFraction(this)
		is Byte -> ImpreciseFraction(this)
		is Short -> ImpreciseFraction(this)
		is Long -> ImpreciseFraction(this)
		is Fraction -> this.toImpreciseFraction()
		else -> throw ClassCastException("Can not turn $this into ImpreciseFraction")
	}
}

fun BigInteger(tag: Tag?): BigInteger {
	if (tag !is ByteArrayTag)
		return BigInteger.ZERO

	return BigInteger(tag.asByteArray)
}

fun BigInteger.serializeNBT(): ByteArrayTag {
	return ByteArrayTag(toByteArray())
}

fun FriendlyByteBuf.writeBigInteger(value: BigInteger) {
	writeByteArray(value.toByteArray())
}

fun FriendlyByteBuf.readBigInteger(byteLimit: Int = 128) = BigInteger(readByteArray(byteLimit))

operator fun IItemHandler.get(index: Int): ItemStack = getStackInSlot(index)

operator fun JsonObject.set(s: String, value: JsonElement) = add(s, value)
operator fun JsonObject.set(s: String, value: String) = add(s, JsonPrimitive(value))
operator fun JsonObject.set(s: String, value: Number) = add(s, JsonPrimitive(value))
operator fun JsonObject.set(s: String, value: Boolean) = add(s, JsonPrimitive(value))

fun <T> LazyOptional<T>.orNull(): T? {
	if (!isPresent) {
		return null
	}

	return resolve().orElse(null)
}

fun <T> LazyOptional<T>.orThrow(): T {
	if (!isPresent) {
		throw IllegalStateException("Capability was expected to be not null")
	}

	return resolve().get() ?: throw IllegalStateException("Capability was expected to be not null")
}

inline fun <T> LazyOptional<T>.ifPresentK(lambda: (T) -> Unit) {
	if (isPresent) {
		val value = resolve().orElse(null) ?: throw IllegalStateException("Capability was expected to be not null")
		lambda.invoke(value)
	}
}

val ItemStack.tagNotNull: CompoundTag get() = orCreateTag

inline var Entity.position: Vec3
	get() = position()
	set(value) { setPos(value) }

inline val <reified T : Enum<T>> T.next: T get() {
	val values = enumValues<T>()
	val next = (ordinal + 1) % values.size
	return values[next]
}

inline val <reified T : Enum<T>> T.prev: T get() {
	val values = enumValues<T>()
	var next = ordinal - 1

	if (next < 0) {
		next = values.size - 1
	}

	return values[next]
}

fun <T : Enum<T>> T.next(values: Array<out T>): T {
	val next = (ordinal + 1) % values.size
	return values[next]
}

fun <T : Enum<T>> T.prev(values: Array<out T>): T {
	var next = ordinal - 1

	if (next < 0) {
		next = values.size - 1
	}

	return values[next]
}

inline fun <T> ImmutableList(size: Int, initializer: (index: Int) -> T): ImmutableList<T> {
	require(size >= 0) { "Invalid list size $size" }

	return when (size) {
		0 -> ImmutableList.of()
		1 -> ImmutableList.of(initializer(0))
		else -> ImmutableList.Builder<T>().let {
			for (i in 0 until size) {
				it.add(initializer(i))
			}

			it.build()
		}
	}
}

fun <T> IForgeRegistry<T>.getID(value: T): Int {
	return (this as ForgeRegistry<T>).getID(value)
}

fun <T> IForgeRegistry<T>.getValue(index: Int): T? {
	return (this as ForgeRegistry<T>).getValue(index)
}

fun IForgeRegistry<*>.getID(value: ResourceLocation): Int {
	return (this as ForgeRegistry<*>).getID(value)
}

fun FriendlyByteBuf.writeItemType(value: Item) {
	writeInt(ForgeRegistries.ITEMS.getID(value))
}

fun FriendlyByteBuf.readItemType(): Item? {
	return ForgeRegistries.ITEMS.getValue(readInt())
}

operator fun <T : Comparable<T>> StateHolder<*, *>.get(property: Property<T>): T {
	return getValue(property)
}

operator fun <S : StateHolder<*, *>, T : Comparable<T>> S.set(property: Property<T>, value: T): S {
	setValue(property, value)
	return this
}

fun <T> List<T>.toImmutableList(): List<T> {
	return ImmutableList.copyOf(this)
}

operator fun <R> (() -> R).getValue(thisRef: Any, property: KProperty<*>): R {
	return invoke()
}

fun BlockState.getExplosionResistance(level: BlockGetter, pos: BlockPos): Float {
	return try {
		getExplosionResistance(level, pos, null)
	} catch (err: NullPointerException) {
		block.explosionResistance
		// Потому что возможно какой-либо мод не ожидает что Explosion == null
		// особенно учитывая что интерфейс IForgeBlock не имеет @ParamsAreNonnullByDefault
		// и аргумент не помечен как @Nullable
		// тем самым имеет тип Explosion! который указывается как Explosion? .. Explosion!!
	} catch (err: IllegalArgumentException) {
		block.explosionResistance
	}
}

fun <E> MutableCollection<E>.addAll(elements: Iterator<E>) {
	for (item in elements) {
		add(item)
	}
}

fun <E> MutableCollection<E>.addAll(elements: Stream<out E>) {
	for (item in elements) {
		add(item)
	}
}

fun <E> Iterable<E>.stream(): Stream<out E> {
	return StreamSupport.stream(this.spliterator(), false)
}

fun <E> Iterator<E>.stream(): Stream<out E> {
	return StreamSupport.stream(Spliterators.spliteratorUnknownSize(this, 0), false)
}

val Ingredient.isActuallyEmpty: Boolean get() = ForgeHooks.hasNoElements(this)

fun Entity.genericPositions(): Collection<Vector> {
	return listOf(
		position,
		eyePosition,
		boundingBox.center
	)
}

fun ResourceLocation.toUUID(): UUID {
	return toString().toUUID()
}

fun String.toUUID(): UUID {
	return UUID.nameUUIDFromBytes(toByteArray(Charsets.UTF_8))
}

fun <T> Collection<T>.probablyParallelStream(): Stream<out T> {
	if (size >= 400) {
		return parallelStream()
	}

	return stream()
}

fun <T> Array<T>.stream(): Stream<T> = Arrays.stream(this)

fun <T> Stream<T?>.filterNotNull(): Stream<T> {
	return filter { it != null } as Stream<T>
}

inline fun <reified T> Stream<*>.filterIsInstance(): Stream<T> {
	return filter { it is T } as Stream<T>
}
