package ru.dbotthepony.mc.otm.network

import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.io.FastByteArrayInputStream
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ReferenceArraySet
import net.minecraft.world.item.ItemStack
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.secondTime
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.lang.ref.WeakReference
import java.math.BigDecimal
import java.util.*
import java.util.function.Consumer
import java.util.function.Supplier
import kotlin.ConcurrentModificationException
import kotlin.collections.ArrayList
import kotlin.collections.HashMap
import kotlin.properties.ReadOnlyProperty
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KMutableProperty0
import kotlin.reflect.KProperty
import kotlin.reflect.KProperty0

interface FieldAccess<V> {
	fun read(): V
	fun write(value: V)
}

fun interface FieldGetter<V> {
	fun invoke(field: FieldAccess<V>): V
}

fun interface FieldSetter<V> {
	fun invoke(value: V, field: FieldAccess<V>, setByRemote: Boolean)
}

sealed interface IField<V> : ReadOnlyProperty<Any, V>, Supplier<V>, () -> V {
	fun observe(): Boolean
	fun markDirty()
	fun markDirty(endpoint: FieldSynchronizer.Endpoint)
	val value: V
	val name: String

	fun write(stream: DataOutputStream, endpoint: FieldSynchronizer.Endpoint)
	fun read(stream: DataInputStream, payloadSize: Int) {
		read(stream)
	}

	fun read(stream: DataInputStream)

	override fun getValue(thisRef: Any, property: KProperty<*>): V {
		return this.value
	}

	override fun get(): V {
		return value
	}

	override fun invoke(): V {
		return value
	}
}

sealed interface IMutableField<V> : IField<V>, ReadWriteProperty<Any, V>, Consumer<V> {
	override var value: V

	override fun setValue(thisRef: Any, property: KProperty<*>, value: V) {
		this.value = value
	}

	override fun accept(t: V) {
		value = t
	}

	override fun getValue(thisRef: Any, property: KProperty<*>): V {
		return super.getValue(thisRef, property)
	}
}

data class MapChangeset<out K, out V>(
	val action: MapAction,
	val key: K?,
	val value: V?
)

enum class MapAction {
	CLEAR, ADD, REMOVE
}

/**
 * Universal one-to-many value synchronizer, allowing to synchronize values from server to client
 * anywhere, where input/output streams are supported
 */
@Suppress("unused")
class FieldSynchronizer(private val callback: Runnable, private val alwaysCallCallback: Boolean) {
	constructor() : this(Runnable {}, false)
	constructor(callback: Runnable) : this(callback, false)

	private val fields = ArrayList<AbstractField<*>>(0)
	private val observers = ArrayList<AbstractField<*>>(0)

	private var nextFieldID = 0

	/**
	 * returns dummy field name
	 */
	fun nextFieldName() = "___field_${nextFieldID++}"
	private var mappingVersion = 0

	val hasObservers: Boolean get() = observers.isNotEmpty()
	val isEmpty: Boolean get() = fields.isEmpty()
	val isNotEmpty: Boolean get() = fields.isNotEmpty()

	var hasChanges: Boolean = false
		private set(value) {
			if (value != field) {
				field = value

				if (value && !alwaysCallCallback) {
					callback.run()
				}
			}

			if (alwaysCallCallback && value) {
				callback.run()
			}
		}

	fun markClean() {
		hasChanges = false
	}

	fun byte(getter: () -> Byte, name: String = nextFieldName()) = ComputedField(getter, ByteValueCodec, name)
	fun bool(getter: () -> Boolean, name: String = nextFieldName()) = ComputedField(getter, BooleanValueCodec, name)
	fun short(getter: () -> Short, name: String = nextFieldName()) = ComputedField(getter, ShortValueCodec, name)
	fun long(getter: () -> Long, name: String = nextFieldName()) = ComputedField(getter, VarLongValueCodec, name)
	fun fixedLong(getter: () -> Long, name: String = nextFieldName()) = ComputedField(getter, LongValueCodec, name)
	fun float(getter: () -> Float, name: String = nextFieldName()) = ComputedField(getter, FloatValueCodec, name)
	fun double(getter: () -> Double, name: String = nextFieldName()) = ComputedField(getter, DoubleValueCodec, name)
	fun uuid(getter: () -> UUID, name: String = nextFieldName()) = ComputedField(getter, UUIDValueCodec, name)
	fun int(getter: () -> Int, name: String = nextFieldName()) = ComputedField(getter, VarIntValueCodec, name)
	fun fixedInt(getter: () -> Int, name: String = nextFieldName()) = ComputedField(getter, IntValueCodec, name)
	fun fraction(getter: () -> ImpreciseFraction, name: String = nextFieldName()) = ComputedField(getter, ImpreciseFractionValueCodec, name)
	fun bigDecimal(getter: () -> BigDecimal, name: String = nextFieldName()) = ComputedField(getter, BigDecimalValueCodec, name)
	fun item(getter: () -> ItemStack, name: String = nextFieldName()) = ComputedField(getter, ItemStackValueCodec, name)

	fun byte(getter: KProperty0<Byte>, name: String = nextFieldName()) = ComputedField(getter, ByteValueCodec, name)
	fun bool(getter: KProperty0<Boolean>, name: String = nextFieldName()) = ComputedField(getter, BooleanValueCodec, name)
	fun short(getter: KProperty0<Short>, name: String = nextFieldName()) = ComputedField(getter, ShortValueCodec, name)
	fun long(getter: KProperty0<Long>, name: String = nextFieldName()) = ComputedField(getter, VarLongValueCodec, name)
	fun fixedLong(getter: KProperty0<Long>, name: String = nextFieldName()) = ComputedField(getter, LongValueCodec, name)
	fun float(getter: KProperty0<Float>, name: String = nextFieldName()) = ComputedField(getter, FloatValueCodec, name)
	fun double(getter: KProperty0<Double>, name: String = nextFieldName()) = ComputedField(getter, DoubleValueCodec, name)
	fun uuid(getter: KProperty0<UUID>, name: String = nextFieldName()) = ComputedField(getter, UUIDValueCodec, name)
	fun int(getter: KProperty0<Int>, name: String = nextFieldName()) = ComputedField(getter, VarIntValueCodec, name)
	fun fixedInt(getter: KProperty0<Int>, name: String = nextFieldName()) = ComputedField(getter, IntValueCodec, name)
	fun fraction(getter: KProperty0<ImpreciseFraction>, name: String = nextFieldName()) = ComputedField(getter, ImpreciseFractionValueCodec, name)
	fun bigDecimal(getter: KProperty0<BigDecimal>, name: String = nextFieldName()) = ComputedField(getter, BigDecimalValueCodec, name)
	fun item(getter: KProperty0<ItemStack>, name: String = nextFieldName()) = ComputedField(getter, ItemStackValueCodec, name)

	fun byte(getter: Supplier<Byte>, name: String = nextFieldName()) = ComputedField(getter::get, ByteValueCodec, name)
	fun bool(getter: Supplier<Boolean>, name: String = nextFieldName()) = ComputedField(getter::get, BooleanValueCodec, name)
	fun short(getter: Supplier<Short>, name: String = nextFieldName()) = ComputedField(getter::get, ShortValueCodec, name)
	fun long(getter: Supplier<Long>, name: String = nextFieldName()) = ComputedField(getter::get, VarLongValueCodec, name)
	fun fixedLong(getter: Supplier<Long>, name: String = nextFieldName()) = ComputedField(getter::get, LongValueCodec, name)
	fun float(getter: Supplier<Float>, name: String = nextFieldName()) = ComputedField(getter::get, FloatValueCodec, name)
	fun double(getter: Supplier<Double>, name: String = nextFieldName()) = ComputedField(getter::get, DoubleValueCodec, name)
	fun uuid(getter: Supplier<UUID>, name: String = nextFieldName()) = ComputedField(getter::get, UUIDValueCodec, name)
	fun int(getter: Supplier<Int>, name: String = nextFieldName()) = ComputedField(getter::get, VarIntValueCodec, name)
	fun fixedInt(getter: Supplier<Int>, name: String = nextFieldName()) = ComputedField(getter::get, IntValueCodec, name)
	fun fraction(getter: Supplier<ImpreciseFraction>, name: String = nextFieldName()) = ComputedField(getter::get, ImpreciseFractionValueCodec, name)
	fun bigDecimal(getter: Supplier<BigDecimal>, name: String = nextFieldName()) = ComputedField(getter::get, BigDecimalValueCodec, name)
	fun item(getter: Supplier<ItemStack>, name: String = nextFieldName()) = ComputedField(getter::get, ItemStackValueCodec, name)

	fun <T : Enum<T>> enum(type: Class<T>, getter: () -> T, name: String = nextFieldName()) = ComputedField(getter, EnumValueCodec(type), name)
	inline fun <reified T : Enum<T>> enum(noinline getter: () -> T, name: String = nextFieldName()) = ComputedField(getter, EnumValueCodec(T::class.java), name)

	fun <T : Enum<T>> enum(type: Class<T>, getter: KProperty0<T>, name: String = nextFieldName()) = ComputedField(getter, EnumValueCodec(type), name)
	inline fun <reified T : Enum<T>> enum(getter: KProperty0<T>, name: String = nextFieldName()) = ComputedField(getter, EnumValueCodec(T::class.java), name)

	fun <T : Enum<T>> enum(type: Class<T>, getter: Supplier<T>, name: String = nextFieldName()) = ComputedField(getter::get, EnumValueCodec(type), name)
	inline fun <reified T : Enum<T>> enum(getter: Supplier<T>, name: String = nextFieldName()) = ComputedField(getter::get, EnumValueCodec(T::class.java), name)

	fun byte(
		value: Byte = 0,
		getter: FieldGetter<Byte>? = null,
		setter: FieldSetter<Byte>? = null,
		name: String = nextFieldName(),
	): Field<Byte> {
		return Field(value, ByteValueCodec, getter, setter, name = name)
	}

	fun bool(
		value: Boolean = false,
		getter: FieldGetter<Boolean>? = null,
		setter: FieldSetter<Boolean>? = null,
		name: String = nextFieldName(),
	): Field<Boolean> {
		return Field(value, BooleanValueCodec, getter, setter, name = name)
	}

	fun short(
		value: Short = 0,
		getter: FieldGetter<Short>? = null,
		setter: FieldSetter<Short>? = null,
		name: String = nextFieldName(),
	): Field<Short> {
		return Field(value, ShortValueCodec, getter, setter, name = name)
	}

	fun long(
		value: Long = 0L,
		getter: FieldGetter<Long>? = null,
		setter: FieldSetter<Long>? = null,
		name: String = nextFieldName(),
	): Field<Long> {
		return Field(value, VarLongValueCodec, getter, setter, name = name)
	}

	fun fixedLong(
		value: Long = 0L,
		getter: FieldGetter<Long>? = null,
		setter: FieldSetter<Long>? = null,
		name: String = nextFieldName(),
	): Field<Long> {
		return Field(value, LongValueCodec, getter, setter, name = name)
	}

	fun float(
		value: Float = 0f,
		getter: FieldGetter<Float>? = null,
		setter: FieldSetter<Float>? = null,
		name: String = nextFieldName(),
	): Field<Float> {
		return Field(value, FloatValueCodec, getter, setter, name = name)
	}

	fun double(
		value: Double = 0.0,
		getter: FieldGetter<Double>? = null,
		setter: FieldSetter<Double>? = null,
		name: String = nextFieldName(),
	): Field<Double> {
		return Field(value, DoubleValueCodec, getter, setter, name = name)
	}

	fun uuid(
		value: UUID = UUID(0L, 0L),
		getter: FieldGetter<UUID>? = null,
		setter: FieldSetter<UUID>? = null,
		name: String = nextFieldName(),
	): Field<UUID> {
		return Field(value, UUIDValueCodec, getter, setter, name = name)
	}

	fun int(
		value: Int = 0,
		getter: FieldGetter<Int>? = null,
		setter: FieldSetter<Int>? = null,
		name: String = nextFieldName(),
	): Field<Int> {
		return Field(value, VarIntValueCodec, getter, setter, name = name)
	}

	fun fixedInt(
		value: Int = 0,
		getter: FieldGetter<Int>? = null,
		setter: FieldSetter<Int>? = null,
		name: String = nextFieldName(),
	): Field<Int> {
		return Field(value, IntValueCodec, getter, setter, name = name)
	}

	fun fraction(
		value: ImpreciseFraction = ImpreciseFraction.ZERO,
		getter: FieldGetter<ImpreciseFraction>? = null,
		setter: FieldSetter<ImpreciseFraction>? = null,
		name: String = nextFieldName(),
	): Field<ImpreciseFraction> {
		return Field(value, ImpreciseFractionValueCodec, getter, setter, name = name)
	}

	fun bigDecimal(
		value: BigDecimal = BigDecimal.ZERO,
		getter: FieldGetter<BigDecimal>? = null,
		setter: FieldSetter<BigDecimal>? = null,
		name: String = nextFieldName(),
	): Field<BigDecimal> {
		return Field(value, BigDecimalValueCodec, getter, setter, name = name)
	}

	fun <T : Enum<T>> enum(
		type: Class<T>,
		value: T = type.enumConstants[0],
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,
		name: String = nextFieldName(),
	): Field<T> {
		return Field(value, EnumValueCodec(type), getter, setter, name = name)
	}

	fun <T : Enum<T>> enum(
		value: T,
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,
		name: String = nextFieldName(),
	): Field<T> {
		return Field(value, EnumValueCodec(value::class.java), getter, setter, name = name)
	}

	fun item(
		value: ItemStack = ItemStack.EMPTY,
		getter: FieldGetter<ItemStack>? = null,
		setter: FieldSetter<ItemStack>? = null,
		observe: Boolean = true,
		name: String = nextFieldName(),
	): Field<ItemStack> {
		return Field(value, ItemStackValueCodec, getter, setter, isObserver = observe, name = name)
	}

	fun item(
		delegate: KMutableProperty0<ItemStack>,
		name: String = nextFieldName(),
	): ObservedField<ItemStack> {
		return ObservedField(delegate, ItemStackValueCodec, name = name)
	}

	fun <K, V> map(
		keyCodec: IStreamCodec<K>,
		valueCodec: IStreamCodec<V>,
		callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null,
		backingMap: MutableMap<K, V> = HashMap(),
		observingBackingMap: MutableMap<K, V>? = null,
		name: String = nextFieldName(),
	): Map<K, V> {
		return Map(
			keyCodec = keyCodec,
			valueCodec = valueCodec,
			callback = callback,
			backingMap = backingMap,
			observingBackingMap = observingBackingMap,
			name = name,
		)
	}

	private var endpointsMaxCapacity = 1
	private val endpoints = ArrayList<WeakReference<Endpoint>>(1)
	val defaultEndpoint = Endpoint()

	private var nextEndpointsCleanup = secondTime

	private fun notifyEndpoints(dirtyField: AbstractField<*>) {
		hasChanges = true

		forEachEndpoint {
			it.addDirtyField(dirtyField)
		}
	}

	private inline fun forEachEndpoint(execute: (Endpoint) -> Unit) {
		nextEndpointsCleanup = secondTime + 60

		synchronized(endpoints) {
			val iterator = endpoints.listIterator()

			for (value in iterator) {
				val endpoint = value.get()

				if (endpoint == null) {
					iterator.remove()
				} else {
					execute.invoke(endpoint)
				}
			}

			if (endpoints.size < endpointsMaxCapacity / 2) {
				endpoints.trimToSize()
				endpointsMaxCapacity = endpoints.size
			}
		}
	}

	inner class Endpoint {
		init {
			synchronized(endpoints) {
				endpoints.add(WeakReference(this))
				endpointsMaxCapacity = endpointsMaxCapacity.coerceAtLeast(endpoints.size)

				if (secondTime >= nextEndpointsCleanup) {
					nextEndpointsCleanup = secondTime + 60

					val iterator = endpoints.listIterator()

					for (value in iterator) {
						if (value.get() == null) {
							iterator.remove()
						}
					}

					if (endpoints.size < endpointsMaxCapacity / 2) {
						endpoints.trimToSize()
						endpointsMaxCapacity = endpoints.size
					}
				}
			}
		}

		private val dirtyFields = ReferenceArraySet<AbstractField<*>>(fields.size)

		// use LinkedList because it is ensured memory is freed on LinkedList#clear
		private val mapBacklogs = Reference2ObjectOpenHashMap<Map<*, *>, LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

		var unused: Boolean = false
			private set

		private var mappingVersion = -1

		fun invalidateMappings() {
			mappingVersion = -1
		}

		fun markUnused() {
			require(this === defaultEndpoint) { "This is not a default endpoint" }
			if (unused) return
			unused = true
			mapBacklogs.clear()
			dirtyFields.clear()

			val iterator = endpoints.listIterator()

			for (value in iterator) {
				if (value.get() === this) {
					iterator.remove()
				}
			}
		}

		init {
			markDirty()
		}

		fun markDirty() {
			for (field in fields) {
				field.markDirty(this)
			}
		}

		internal fun addDirtyField(field: AbstractField<*>) {
			if (unused) {
				return
			}

			dirtyFields.add(field)
		}

		internal fun <K, V> getMapBacklog(map: Map<K, V>): LinkedList<Pair<Any?, (DataOutputStream) -> Unit>> {
			if (unused) {
				return LinkedList()
			}

			return mapBacklogs.computeIfAbsent(map, Reference2ObjectFunction {
				LinkedList()
			})
		}

		fun collectNetworkPayload(): FastByteArrayOutputStream? {
			if (unused || dirtyFields.isEmpty()) {
				return null
			}

			val stream = FastByteArrayOutputStream()

			if (mappingVersion != this@FieldSynchronizer.mappingVersion) {
				stream.write(1)

				for (field in fields) {
					check(field.id > 0) { "This should never happen: $field maps to invalid ID: ${field.id}!" }
					stream.writeVarIntLE(field.id)
					val bytes = field.name.toByteArray(Charsets.UTF_8)
					stream.writeVarIntLE(bytes.size)
					stream.write(bytes)
				}

				stream.writeVarIntLE(0)
				mappingVersion = this@FieldSynchronizer.mappingVersion
			} else {
				stream.write(0)
			}

			for (field in dirtyFields) {
				val id = field.id
				check(id > 0) { "This should never happen: $field maps to invalid ID: $id!" }
				stream.writeVarIntLE(id)

				val innerStream = FastByteArrayOutputStream()
				val dataStream = DataOutputStream(innerStream)

				field.write(dataStream, this)

				stream.writeVarIntLE(innerStream.length)
				stream.write(innerStream.array, 0, innerStream.length)
			}

			dirtyFields.clear()
			stream.write(0)

			return stream
		}
	}

	private val boundEndpoints = WeakHashMap<Any, Endpoint>()

	fun computeEndpointFor(obj: Any): Endpoint {
		return boundEndpoints.computeIfAbsent(obj) { Endpoint() }
	}

	fun removeEndpointFor(obj: Any): Endpoint? {
		return boundEndpoints.remove(obj)
	}

	fun endpointFor(obj: Any): Endpoint? {
		return boundEndpoints[obj]
	}

	@Suppress("LeakingThis")
	abstract inner class AbstractField<V>(
		final override val name: String = nextFieldName(),
	) : IField<V> {
		var id: Int = fields.size + 1

		init {
			check(!fields.any { it.name == name }) { "Duplicate field name $name" }
			fields.add(this)
			mappingVersion++
		}
	}

	inner class Field<V>(
		private var field: V,
		private val codec: IStreamCodec<V>,
		private val getter: FieldGetter<V>? = null,
		private val setter: FieldSetter<V>? = null,
		isObserver: Boolean = false,
		name: String = nextFieldName(),
	) : AbstractField<V>(name), IMutableField<V> {
		private var remote: V = codec.copy(field)

		init {
			if (isObserver) {
				observers.add(this)
			}
		}

		private var isDirty = false

		private val access = object : FieldAccess<V> {
			override fun read(): V {
				return field
			}

			override fun write(value: V) {
				if (!isDirty && !codec.compare(remote, value)) {
					notifyEndpoints(this@Field)
					isDirty = true
				}

				this@Field.field = value
			}
		}

		override fun observe(): Boolean {
			if (!isDirty && !codec.compare(remote, field)) {
				notifyEndpoints(this@Field)
				isDirty = true
			}

			return isDirty
		}

		override var value: V
			get() {
				val getter = this.getter

				if (getter != null) {
					return getter.invoke(access)
				}

				return this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
					return
				}

				if (this.field == value) {
					return
				}

				if (!isDirty && !codec.compare(remote, value)) {
					notifyEndpoints(this@Field)
					isDirty = true
				}

				this.field = value
			}

		override fun markDirty() {
			notifyEndpoints(this@Field)
			isDirty = true
		}

		override fun markDirty(endpoint: Endpoint) {
			endpoint.addDirtyField(this)
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			codec.write(stream, field)
			isDirty = false
			remote = codec.copy(field)
		}

		override fun read(stream: DataInputStream) {
			val value = codec.read(stream)
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
				return
			}

			this.field = value
		}
	}

	inner class ComputedField<V>(
		private val getter: () -> V,
		private val codec: IStreamCodec<V>,
		name: String = nextFieldName(),
	) : AbstractField<V>(name), IField<V> {
		private var remote: V? = null
		private var clientValue: V? = null
		private var isDirty = false

		init {
			observers.add(this)
		}

		override fun observe(): Boolean {
			if (!isDirty && (remote == null || !codec.compare(remote ?: throw ConcurrentModificationException(), value))) {
				notifyEndpoints(this)
				isDirty = true
				remote = codec.copy(value)
			}

			return isDirty
		}

		override fun markDirty() {
			notifyEndpoints(this)
			isDirty = true
		}

		override fun markDirty(endpoint: Endpoint) {
			endpoint.addDirtyField(this)
		}

		override val value: V
			get() = clientValue ?: getter.invoke()

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			val value = value
			codec.write(stream, value)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			clientValue = codec.read(stream)
		}
	}

	inner class ObservedField<V> : AbstractField<V>, IMutableField<V> {
		private val codec: IStreamCodec<V>
		private val getter: () -> V
		private val setter: (V) -> Unit
		private var remote: V

		override var value: V
			get() = getter.invoke()
			set(value) { setter.invoke(value) }

		constructor(field: KMutableProperty0<V>, codec: IStreamCodec<V>, name: String = nextFieldName()) : super(name) {
			this.codec = codec
			getter = field::get
			setter = field::set
			remote = codec.copy(value)
		}

		constructor(getter: () -> V, setter: (V) -> Unit, codec: IStreamCodec<V>, name: String = nextFieldName()) : super(name) {
			this.codec = codec
			this.getter = getter
			this.setter = setter
			remote = codec.copy(value)
		}

		private var isDirty = false

		init {
			observers.add(this)
		}

		override fun observe(): Boolean {
			if (!isDirty && !codec.compare(remote, value)) {
				notifyEndpoints(this)
				isDirty = true
				remote = codec.copy(value)
			}

			return isDirty
		}

		override fun markDirty() {
			notifyEndpoints(this)
			isDirty = true
		}

		override fun markDirty(endpoint: Endpoint) {
			endpoint.addDirtyField(this)
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			val value = value
			codec.write(stream, value)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			this.value = codec.read(stream)
		}
	}

	inner class Map<K, V>(
		private val keyCodec: IStreamCodec<K>,
		private val valueCodec: IStreamCodec<V>,
		private val backingMap: MutableMap<K, V>,
		private val observingBackingMap: MutableMap<K, V>? = null,
		private val callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null,
		name: String = nextFieldName(),
	) : AbstractField<MutableMap<K, V>>(name), IField<MutableMap<K, V>> {
		private var isDirty = false
		private var sentAllValues = false
		private var isRemote = false

		init {
			if (observingBackingMap != null)
				observers.add(this)
		}

		private fun pushBacklog(key: Any?, value: (DataOutputStream) -> Unit) {
			forEachEndpoint {
				val list = it.getMapBacklog(this)
				val iterator = list.listIterator()

				for (pair in iterator) {
					if (pair.first == key) {
						iterator.remove()
					}
				}

				list.addLast(key to value)
			}
		}

		private fun clearBacklog() {
			forEachEndpoint {
				it.getMapBacklog(this).clear()
			}
		}

		override fun observe(): Boolean {
			if (isRemote) {
				return false
			}

			val observingBackingMap = observingBackingMap

			if (observingBackingMap != null) {
				for ((key, value) in backingMap) {
					val remoteValue = observingBackingMap[key] ?: throw ConcurrentModificationException("Backing map of $this was modified externally, or $value missed a modification")

					if (!valueCodec.compare(value, remoteValue)) {
						val valueCopy = valueCodec.copy(value)

						pushBacklog(key) {
							it.write(MapAction.ADD.ordinal + 1)
							keyCodec.write(it, key)
							valueCodec.write(it, valueCopy)
						}

						observingBackingMap[key] = valueCopy

						if (!isDirty) {
							notifyEndpoints(this)
							isDirty = true
						}
					}
				}
			}

			return isDirty
		}

		override fun markDirty() {
			if (isRemote) {
				return
			}

			if (endpoints.isEmpty()) {
				val observingBackingMap = observingBackingMap ?: return

				for ((key, value) in backingMap)
					observingBackingMap[key] = valueCodec.copy(value)

				return
			}

			isDirty = true
			val backlogs = LinkedList<LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

			forEachEndpoint {
				it.addDirtyField(this)
				val value = it.getMapBacklog(this)
				backlogs.add(value)
				value.clear()
				value.add(null to ClearBacklogEntry)
			}

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				val action = { it: DataOutputStream ->
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				for (backlog in backlogs) {
					backlog.add(key to action)
				}

				observingBackingMap?.put(key, valueCopy)
			}
		}

		override fun markDirty(endpoint: Endpoint) {
			if (isRemote) {
				return
			}

			val backlog = endpoint.getMapBacklog(this)

			backlog.clear()
			backlog.add(null to ClearBacklogEntry)

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				backlog.add(key to {
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				})
			}
		}

		override val value: MutableMap<K, V> = object : ProxiedMap<K, V>(backingMap) {
			override fun onClear() {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				observingBackingMap?.clear()

				forEachEndpoint { endpoint ->
					endpoint.getMapBacklog(this@Map).also {
						it.clear()
						it.add(null to ClearBacklogEntry)
					}
				}

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}

			override fun onValueAdded(key: K, value: V) {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				val valueCopy = valueCodec.copy(value)

				pushBacklog(key) {
					@Suppress("BlockingMethodInNonBlockingContext") // false positive
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				observingBackingMap?.put(key, valueCopy)

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}

			override fun onValueRemoved(key: K, value: V) {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				val keyCopy = keyCodec.copy(key)

				pushBacklog(key) {
					@Suppress("BlockingMethodInNonBlockingContext") // false positive
					it.write(MapAction.REMOVE.ordinal + 1)
					keyCodec.write(it, keyCopy)
				}

				observingBackingMap?.remove(key)

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			sentAllValues = false
			isDirty = false

			val iterator = endpoint.getMapBacklog(this).listIterator()

			for (entry in iterator) {
				entry.second.invoke(stream)
				iterator.remove()
			}

			stream.write(0)
		}

		override fun read(stream: DataInputStream) {
			if (!isRemote) {
				isRemote = true
				clearBacklog()
				observingBackingMap?.clear()
			}

			isDirty = false

			val changeset = LinkedList<MapChangeset<K, V>>()
			var readAction = stream.read() - 1

			while (readAction != -1) {
				if (readAction >= MapActionList.size) {
					throw IndexOutOfBoundsException("Unknown map action with ID $readAction")
				}

				when (MapActionList[readAction]) {
					MapAction.CLEAR -> {
						backingMap.clear()
						changeset.add(ClearMapChangeset)
					}

					MapAction.ADD -> {
						val key = keyCodec.read(stream)
						val value = valueCodec.read(stream)
						backingMap[key] = value
						changeset.add(MapChangeset(MapAction.ADD, key, value))
					}

					MapAction.REMOVE -> {
						val key = keyCodec.read(stream)
						backingMap.remove(key)
						changeset.add(MapChangeset(MapAction.REMOVE, key, null))
					}
				}

				readAction = stream.read() - 1
			}

			if (changeset.size != 0) {
				callback?.invoke(changeset)
			}
		}
	}

	/**
	 * marks all fields dirty, invalidates mappings for each endpoint
	 */
	fun invalidate() {
		for (field in fields) {
			field.markDirty()
		}

		forEachEndpoint {
			it.invalidateMappings()
		}
	}

	/**
	 * Observe changes of all fields with backing computation lambda
	 */
	fun observe(): Boolean {
		var changes = false

		if (observers.isNotEmpty()) {
			for (field in observers) {
				changes = field.observe() || changes
			}
		}

		return changes
	}

	/**
	 * [defaultEndpoint]#collectNetworkPayload
	 */
	fun collectNetworkPayload(): FastByteArrayOutputStream? {
		check(!defaultEndpoint.unused) { "Default endpoint is not used" }
		observe()
		val values = defaultEndpoint.collectNetworkPayload()
		markClean()
		return values
	}

	private val idToField = Int2ObjectAVLTreeMap<IField<*>>()
	private val missingFields = ObjectArraySet<String>()
	private val missingFieldsMap = Int2ObjectAVLTreeMap<String>()

	fun applyNetworkPayload(stream: InputStream): Int {
		if (stream.read() > 0) {
			idToField.clear()
			missingFieldsMap.clear()

			var fieldId = stream.readVarIntLE()

			while (fieldId != 0) {
				val size = stream.readVarIntLE()
				val nameBytes = ByteArray(size)
				stream.read(nameBytes)
				val name = String(nameBytes, Charsets.UTF_8)

				val findField = fields.firstOrNull { it.name == name }

				if (findField == null) {
					if (missingFields.add(name)) {
						LOGGER.error("Unable to find field $name in $this, expect issues!")
					}

					missingFieldsMap[fieldId] = name
				} else {
					idToField[fieldId] = findField
					findField.id = fieldId
				}

				fieldId = stream.readVarIntLE()
			}
		}

		var fieldId = stream.readVarIntLE()
		var i = 0

		while (fieldId != 0) {
			val field = idToField[fieldId]
			val payloadSize = stream.readVarIntLE()

			if (field == null) {
				LOGGER.error("Unable to read field $fieldId (${missingFieldsMap[fieldId]}) because we don't know anything about it! Skipping $payloadSize bytes", IllegalStateException("Unknown field $fieldId"))
				stream.skipNBytes(payloadSize.toLong())
				continue
			}

			val bytes = ByteArray(payloadSize)
			stream.read(bytes)
			field.read(DataInputStream(FastByteArrayInputStream(bytes)), payloadSize)
			fieldId = stream.readVarIntLE()
			i++
		}

		return i
	}

	companion object {
		private val ClearBacklogEntry = { stream: DataOutputStream -> stream.write(MapAction.CLEAR.ordinal + 1) }
		private val MapActionList = MapAction.values()
		private val ClearMapChangeset = MapChangeset(MapAction.CLEAR, null, null)
		private val LOGGER = LogManager.getLogger()
	}
}
